package de.bsvrz.buv.plugin.darstellung.dialogs;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.Platform;
import org.eclipse.draw2d.ColorConstants;
import org.eclipse.gef.commands.CommandStack;
import org.eclipse.jface.viewers.CheckStateChangedEvent;
import org.eclipse.jface.viewers.CheckboxTableViewer;
import org.eclipse.jface.viewers.ICheckStateListener;
import org.eclipse.jface.viewers.IColorProvider;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerDropAdapter;
import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.DND;
import org.eclipse.swt.dnd.DragSourceAdapter;
import org.eclipse.swt.dnd.DragSourceEvent;
import org.eclipse.swt.dnd.DropTargetEvent;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.dnd.TransferData;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbenchPart;

import com.bitctrl.lib.eclipse.dialogs.ToolWindow;
import com.bitctrl.lib.eclipse.emf.gef.commands.MoveCommand;
import com.bitctrl.lib.eclipse.emf.gef.commands.SetCommand;

import de.bsvrz.buv.plugin.darstellung.model.Darstellung;
import de.bsvrz.buv.plugin.darstellung.model.DarstellungPackage;
import de.bsvrz.buv.plugin.darstellung.model.Ebene;
import de.bsvrz.buv.plugin.darstellung.util.DarstellungIcons;

/**
 * Ein schwebendes Fenster, in welchen die Ebenen ein- und ausgeblenden werden
 * knnen.
 * 
 * @author BitCtrl Systems GmbH, Falko Schumann
 * @version $Id$
 * @see IAdaptable
 */
public class EbenenWindow extends ToolWindow {

	private final class LabelProviderExtension extends LabelProvider implements
			IColorProvider {

		@Override
		public String getText(final Object element) {
			return ((Ebene) element).getName();
		}

		@Override
		public Color getForeground(final Object element) {
			final Ebene ebene = (Ebene) element;
			if (ebene.isVisible()) {
				return ColorConstants.black;
			}

			return ColorConstants.gray;
		}

		@Override
		public Color getBackground(final Object element) {
			return null;
		}

	}

	private CheckboxTableViewer viewer;
	private Image ebeneVisible;
	private Image ebeneInvisible;

	{
		setShellStyle(SWT.DIALOG_TRIM | SWT.RESIZE | SWT.TOOL);
	}

	/**
	 * Initialisiert der Ebenemanager.
	 * 
	 * @param part
	 *            der Workbench Part zu dem der Ebenemanager gehrt.
	 * @param partControl
	 *            das Control des Workbench Parts bzw. das Control an dessen
	 *            Ecken der Ebenemanager ausgerichtet wird.
	 */
	public EbenenWindow(final IWorkbenchPart part, final Control partControl) {
		super(part, partControl);
		setCorner(Corner.TopRight);
	}

	@Override
	protected Control createContents(final Composite parent) {
		final Composite container = new Composite(parent, SWT.NONE);
		container.setLayout(new FillLayout());

		if (getDarstellung() != null) {
			createViewer(container);
		} else {
			final Label label = new Label(container, SWT.WRAP);
			label.setText("Keine Darstellung vorhanden.");
			label.setBackground(parent.getDisplay().getSystemColor(
					SWT.COLOR_WHITE));

		}

		return container;
	}

	private void createViewer(final Composite parent) {
		ebeneVisible = DarstellungIcons.StatusEbeneVisible.getImageDescriptor()
				.createImage();
		ebeneInvisible = DarstellungIcons.StatusEbeneInvisible
				.getImageDescriptor().createImage();

		viewer = CheckboxTableViewer.newCheckList(parent, SWT.BORDER
				| SWT.SINGLE);
		viewer.setLabelProvider(new LabelProviderExtension());
		viewer.setContentProvider(new IStructuredContentProvider() {

			@Override
			public Object[] getElements(final Object inputElement) {
				final List<?> list = new ArrayList<Object>(
						(List<?>) inputElement);
				Collections.reverse(list);
				return list.toArray();
			}

			@Override
			public void inputChanged(final Viewer v, final Object oldInput,
					final Object newInput) {
				// tut nix
			}

			@Override
			public void dispose() {
				// tut nix
			}

		});
		viewer.setInput(getDarstellung().getEbenen());
		for (final Ebene e : getDarstellung().getEbenen()) {
			viewer.setChecked(e, e.isVisible());
		}

		// Drag'n'Drop
		final Transfer[] transferTypes = new Transfer[] { ViewerTransfer
				.getInstance() };
		final int operations = DND.DROP_MOVE | DND.DROP_COPY;
		viewer.addDragSupport(operations, transferTypes,
				new DragSourceAdapter() {

					@Override
					public void dragSetData(final DragSourceEvent event) {
						final IStructuredSelection selection = (IStructuredSelection) viewer
								.getSelection();
						final Object firstElement = selection.getFirstElement();

						if (ViewerTransfer.getInstance().isSupportedType(
								event.dataType)) {
							event.data = firstElement;
						}
					}

				});
		viewer.addDropSupport(operations, transferTypes, new ViewerDropAdapter(
				viewer) {

			@Override
			public boolean validateDrop(final Object target,
					final int operation, final TransferData transferType) {
				return true;
			}

			@Override
			public boolean performDrop(final Object data) {
				final Ebene ebene = (Ebene) data;
				final DropTargetEvent event = getCurrentEvent();
				final int location = determineLocation(event);
				final Ebene target = (Ebene) determineTarget(event);
				int targetIndex = getDarstellung().getEbenen().indexOf(target);
				MoveCommand cmd;
				switch (location) {
				case 1:
					// Dropped before the target
					if (targetIndex < getDarstellung().getEbenen().size() - 1) {
						++targetIndex;
					}
					cmd = new MoveCommand(getDarstellung(),
							DarstellungPackage.Literals.DARSTELLUNG__EBENEN,
							ebene, targetIndex);
					break;
				case 2:
					// Dropped after the target
					if (targetIndex > 0) {
						--targetIndex;
					}
					cmd = new MoveCommand(getDarstellung(),
							DarstellungPackage.Literals.DARSTELLUNG__EBENEN,
							ebene, targetIndex);
					break;
				case 3:
					// Dropped on the target
					cmd = new MoveCommand(getDarstellung(),
							DarstellungPackage.Literals.DARSTELLUNG__EBENEN,
							ebene, targetIndex);
					break;
				case 4:
					// Dropped into nothing
				default:
					cmd = null;
				}
				if (cmd != null) {
					getCommandStack().execute(cmd);
					viewer.refresh();
				}

				return true;
			}

		});

		viewer.addCheckStateListener(new ICheckStateListener() {

			@Override
			public void checkStateChanged(final CheckStateChangedEvent event) {
				final Ebene ebene = (Ebene) event.getElement();
				final SetCommand cmd = new SetCommand(ebene,
						DarstellungPackage.Literals.EBENE__VISIBLE, !ebene
								.isVisible());
				getCommandStack().execute(cmd);
				viewer.refresh();
			}

		});

	}

	private Darstellung getDarstellung() {
		Darstellung result = (Darstellung) getWorkbenchPart().getAdapter(
				Darstellung.class);
		if (result == null) {
			result = (Darstellung) Platform.getAdapterManager().getAdapter(
					getWorkbenchPart(), Darstellung.class);
		}

		return result;
	}

	private CommandStack getCommandStack() {
		CommandStack result = (CommandStack) getWorkbenchPart().getAdapter(
				CommandStack.class);
		if (result == null) {
			result = (CommandStack) Platform.getAdapterManager().getAdapter(
					getWorkbenchPart(), CommandStack.class);
		}

		return result;
	}

	@Override
	protected void configureShell(final Shell newShell) {
		super.configureShell(newShell);

		newShell.setText("Ebenen");
		newShell.setSize(200, 400);
	}

	@Override
	public boolean close() {
		if (ebeneVisible != null) {
			ebeneVisible.dispose();
			ebeneVisible = null;
		}
		if (ebeneInvisible != null) {
			ebeneInvisible.dispose();
			ebeneInvisible = null;
		}

		return super.close();
	}

}
