package de.bsvrz.buv.plugin.darstellung.wizards;

import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.PlatformUI;

import com.bitctrl.lib.eclipse.emf.eclipse.model.Named;

import de.bsvrz.buv.plugin.dobj.internal.DObjPlugin;
import de.bsvrz.buv.rw.basislib.einstellungen.EinstellungenWaehler.EinstellungsArt;
import de.bsvrz.buv.rw.bitctrl.eclipse.modell.emf.EMFModellEinstellungen;
import de.bsvrz.buv.rw.bitctrl.eclipse.wizards.EinstellungsArtWizardPage;

/**
 * Basisklasse fr Wizards die {@link EObject}-Objekte importieren, die in den
 * Rahmenwerkseinstellungen gespeichert werden.
 * 
 * @author BitCtrl Systems GmbH, Falko Schumann
 * @version $Id$
 */
public abstract class AbstractImportWizard extends AbstractWorkbenchWizard {

	private final EMFModellEinstellungen einstellungen;

	private AbstractImportWizardPage importPage;
	private EinstellungsArtWizardPage einstellungsArtWizardPage;

	/**
	 * Initialisiert den Wizard.
	 * 
	 * @param eClass
	 *            der Typ der zu importierenden Objekte.
	 */
	public AbstractImportWizard(final EClass eClass,
			final EMFModellEinstellungen<?> einstellungen) {
		super(eClass);
		this.einstellungen = einstellungen;
	}

	@Override
	public void addPages() {
		importPage = getImportWizardPage();
		addPage(importPage);

		einstellungsArtWizardPage = new EinstellungsArtWizardPage(
				EinstellungsArtWizardPage.class.getName());
		final Set<EinstellungsArt> einstelungsArten = new HashSet<EinstellungsArt>();
		for (final Object e : (getSelection()).toList()) {
			if (e instanceof EinstellungsArt) {
				einstelungsArten.add((EinstellungsArt) e);
			}
		}
		einstellungsArtWizardPage.setPreSelection(einstelungsArten);
		addPage(einstellungsArtWizardPage);
	}

	protected abstract AbstractImportWizardPage getImportWizardPage();

	@Override
	public boolean performFinish() {
		final List<String> fileNames = importPage.getFileNames();
		for (final EinstellungsArt einstellungsArt : einstellungsArtWizardPage
				.getEinstellungsArten()) {
			for (final String fileName : fileNames) {
				// Datei lesen
				final ResourceSet resSet = new ResourceSetImpl();
				final File file = new File(fileName);
				final Resource resource = resSet.createResource(URI
						.createURI(file.toURI().toString()));
				try {
					resource.load(null);
				} catch (final IOException ex) {
					final String message = "Die Datei konnte nicht gelesen werden: "
							+ file;
					final IStatus status = DObjPlugin.getDefault().getLogger()
							.error(message, ex);

					ErrorDialog.openError(PlatformUI.getWorkbench()
							.getActiveWorkbenchWindow().getShell(),
							"Fehler beim Import", message, status);

					return false;
				}

				// Dateiinhalt prfen
				final EList<EObject> contents = resource.getContents();
				if (contents.size() != 1) {
					final String message = "Die Datei ist fehlerhaft und kann nicht gelesen werden: "
							+ file;
					final IStatus status = DObjPlugin.getDefault().getLogger()
							.error(message);

					ErrorDialog.openError(PlatformUI.getWorkbench()
							.getActiveWorkbenchWindow().getShell(),
							"Fehler beim Import", message, status);
					return false;
				}
				final Named named = (Named) contents.get(0);

				// Darstellung importieren
				if (einstellungen.getModellEinstellungen(einstellungsArt,
						named.getName()) != null) {
					if (MessageDialog.openQuestion(getShell(), getEClass()
							.getInstanceTypeName() + " berschreiben",
							"Es existiert bereits ein Objekt mit dem selben Namen "
									+ named.getName()
									+ ". Wollen Sie es berschreiben?")) {
						einstellungen.setModellEinstellungen(einstellungsArt,
								named.getName(), named);
						return true;
					}

					return false;
				} else {
					einstellungen.setModellEinstellungen(einstellungsArt,
							named.getName(), named);
				}
			}
		}

		return true;
	}

}
