package de.bsvrz.buv.plugin.dobj.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.eclipse.draw2d.ColorConstants;
import org.eclipse.draw2d.Graphics;
import org.eclipse.draw2d.Shape;
import org.eclipse.draw2d.geometry.PointList;
import org.eclipse.gef.EditPart;
import org.eclipse.gef.EditPartViewer;
import org.eclipse.gef.GraphicalEditPart;
import org.eclipse.jface.viewers.StructuredSelection;

import de.bsvrz.buv.plugin.dobj.model.BitCtrlDoModel;
import de.bsvrz.buv.plugin.dobj.model.DoModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.bitctrl.modell.SystemObjekt;

/**
 * Stellt allgemeine Funktionen fr Darstellungsobjekte bereit.
 * 
 * @author BitCtrl Systems GmbH, Falko Schumann
 * @version $Id$
 */
public final class DobjUtil {

	/**
	 * Konstante fr die Viewer-Property, ob der Edit Part sich in einem Editor
	 * zum Bearbeiten befindet.
	 */
	public static final String PROP_EDITOR = "de.bsvrz.buv.plugin.dobj.editor";

	/**
	 * Konstante fr die Viewer-Property, die den Typ des Editors definiert.
	 */
	public static final String PROP_EDITOR_TYPE = "de.bsvrz.buv.plugin.dobj.editorType";

	/** Konstante fr die Kennzeichnung eines DO-Typ-Editors. */
	public static final int PROP_EDITOR_TYPE_DOTYP = 1;

	/** Konstante fr die Kennzeichnung eines Darstellungs-Editors. */
	public static final int PROP_EDITOR_TYPE_DARSTELLUNG = 2;

	/**
	 * Vergleicht ob zwei {@link PointList}s die gleichen Punkte in der gleichen
	 * Reihenfolge enthalten. Gibt auch <code>true</code> zurck, wenn beide
	 * Parameter <code>null</code> sind.
	 */
	public static boolean equals(final PointList pointList1,
			final PointList pointList2) {
		if (pointList1 == pointList2) {
			return false;
		}

		if (pointList1 == null || pointList2 == null) {
			return false;
		}

		final int[] points1 = pointList1.toIntArray();
		final int[] points2 = pointList2.toIntArray();
		return Arrays.equals(points1, points2);
	}

	/**
	 * Konfiguriert eine Figure fr die Verwendung als Feedback Figure.
	 * 
	 * @param feedbackFigure
	 *            die zu konfigurierende Figure.
	 */
	public static void configureFeedbackFigure(final Shape feedbackFigure) {
		feedbackFigure.setXOR(true);
		feedbackFigure.setForegroundColor(ColorConstants.white);
		feedbackFigure.setBackgroundColor(ColorConstants.black);
		feedbackFigure.setLineStyle(Graphics.LINE_DASH);
	}

	/**
	 * Selektiert alle bergebenen Systemobjekte im Viewer. Im Viewer nicht
	 * vorhandene Systemobjekte werden ignoriert.
	 * 
	 * @param viewer
	 *            der Viewer in dem die Systemobjekte selektiert werden sollen.
	 * @param systemObjects
	 *            die zu selektierenden Systemobjekte ({@link SystemObject} oder
	 *            {@link SystemObjekt}).
	 */
	public static void setSelection(final EditPartViewer viewer,
			final Collection<?> systemObjects) {
		final Map<?, ?> editPartRegistry = viewer.getEditPartRegistry();
		final List<EditPart> editParts = new ArrayList<EditPart>();

		for (final Object o : systemObjects) {
			for (final Entry<?, ?> e : editPartRegistry.entrySet()) {
				if (e.getKey() instanceof DoModel) {
					final DoModel model = (DoModel) e.getKey();
					final EditPart editPart = (EditPart) e.getValue();

					if (o instanceof SystemObject) {
						if (o.equals(model.getSystemObject())) {
							editParts.add(editPart);
						}
					} else if (o instanceof SystemObjekt
							&& model instanceof BitCtrlDoModel<?>) {
						if (o.equals(((BitCtrlDoModel<?>) model)
								.getSystemObjekt())) {
							editParts.add(editPart);
						}
					}
				}
			}
		}

		viewer.setSelection(new StructuredSelection(editParts));
		if (editParts.size() > 0) {
			viewer.reveal(editParts.get(editParts.size() - 1));
		}
	}

	/**
	 * Gibt alle Kinder und Kindeskinder eines Edit Parts zurck.
	 * 
	 * @param editPart
	 *            ein Edit Part.
	 * @return alle untergeordneten Edit Parts.
	 */
	public static Set<EditPart> getAllChildren(final EditPart editPart) {
		final Set<EditPart> allChildren = new HashSet<EditPart>();
		getAllChildren(editPart, allChildren);
		return allChildren;
	}

	private static void getAllChildren(final EditPart editPart,
			final Set<EditPart> allChildren) {
		final List<?> children = editPart.getChildren();
		for (int i = 0; i < children.size(); i++) {
			final GraphicalEditPart child = (GraphicalEditPart) children.get(i);
			allChildren.add(child);
			getAllChildren(child, allChildren);
		}
	}

	private DobjUtil() {
		// utility class
	}

}
